#include "dispatcher.h"

Dispatcher::Dispatcher(QObject *parent) : QObject(parent), m_uploader(this)
{
}

rfbBool Dispatcher::MallocFrameBuffer(rfbClient *client)
{
    Dispatcher *dispatcher = (Dispatcher *) rfbClientGetClientData(client, 0);
    emit dispatcher->resizeFramebuffer(QSize(client->width, client->height));
    return TRUE;
}

void Dispatcher::GotCopyRect(rfbClient *client, int src_x, int src_y, int w, int h, int dest_x, int dest_y)
{
    Dispatcher *dispatcher = (Dispatcher *) rfbClientGetClientData(client, 0);
    dispatcher->uploader().gotCopyRect(src_x, src_y, w, h, dest_x, dest_y);
}

void Dispatcher::GotFillRect(rfbClient *client, int x, int y, int w, int h, uint32_t colour)
{
    Dispatcher *dispatcher = (Dispatcher *) rfbClientGetClientData(client, 0);
    int red = (colour >> client->format.redShift) & client->format.redMax;
    int green = (colour >> client->format.greenShift) & client->format.greenMax;
    int blue = (colour >> client->format.blueShift) & client->format.blueMax;
    dispatcher->uploader().gotFillRect(x, y, w, h, red, green, blue);
}

void Dispatcher::GotBitmap(rfbClient *client, const uint8_t *buffer, int x, int y, int w, int h)
{
    QImage img;
    Dispatcher *dispatcher = (Dispatcher *) rfbClientGetClientData(client, 0);
    switch(dispatcher->bitsPerPixel()) {
    case 8:
        img = QImage(buffer, w, h, QImage::Format_Grayscale8);
        break;
    case 16:
        img = QImage(buffer, w, h, QImage::Format_RGB16);
        break;
    case 32:
        img = QImage(buffer, w, h, QImage::Format_RGBX8888);
        break;
    default:
        emit dispatcher->error();
        return;
    }
    // We have to make a deep copy because libvnc might free the buffer as soon as this function returns
    dispatcher->uploader().gotBitmap(img.copy(), x, y, w, h);
}

rfbBool Dispatcher::GotJpeg(rfbClient *client, const uint8_t *buffer, int length, int x, int y, int w, int h)
{
    Dispatcher *dispatcher = static_cast<Dispatcher *>(rfbClientGetClientData(client, 0));

    QByteArray jpeg;
    jpeg.setRawData((const char*) buffer, length);

    JpegDecoder *decoder = new JpegDecoder(jpeg, x, y, w, h);
    dispatcher->incrementDecoderCount();
    connect(decoder, &JpegDecoder::finished, dispatcher, &Dispatcher::decodingFinished, Qt::QueuedConnection);
    QThreadPool::globalInstance()->start(decoder);

    return TRUE;
}

void Dispatcher::FinishedFrameBufferUpdate(rfbClient *client)
{
    Dispatcher *dispatcher = (Dispatcher *) rfbClientGetClientData(client, 0);
    dispatcher->setFinished(true);
    if(dispatcher->decoderCount() == 0) {
        dispatcher->uploader().finishedUpdate();
        emit dispatcher->updateFramebuffer();
    }
}

void Dispatcher::setMouse(int x, int y, bool left, bool middle, bool right)
{
    if(m_client == NULL)
        return;

    int buttonMask = 0;
    if(left) buttonMask |= rfbButton1Mask;
    if(middle) buttonMask |= rfbButton2Mask;
    if(right) buttonMask |= rfbButton3Mask;
    SendPointerEvent(m_client, x, y, buttonMask);
}

unsigned int Dispatcher::decoderCount() const
{
    return m_decoderCount;
}

unsigned int Dispatcher::incrementDecoderCount()
{
    return ++m_decoderCount;
}

unsigned int Dispatcher::decrementDecoderCount()
{
    return --m_decoderCount;
}

int Dispatcher::bitsPerPixel()
{
    return m_client->format.bitsPerPixel;
}

Uploader &Dispatcher::uploader()
{
    return m_uploader;
}

void Dispatcher::setFinished(bool finished)
{
    m_finished = finished;
}

bool Dispatcher::finished() const
{
    return m_finished;
}

unsigned int Dispatcher::fill_count()
{
    if(m_finished && m_decoderCount == 0)
        return m_uploader.fill_count();
    else
        return 0;
}

unsigned int Dispatcher::copy_count()
{
    if(m_finished && m_decoderCount == 0)
        return m_uploader.copy_count();
    else
        return 0;
}

unsigned int Dispatcher::draw_count()
{
    if(m_finished && m_decoderCount == 0)
        return m_uploader.draw_count();
    else
        return 0;
}

QVector3D Dispatcher::colorMax() const
{
    if(m_client == NULL)
        return QVector3D();

    return QVector3D(m_client->format.redMax, m_client->format.greenMax, m_client->format.blueMax);
}

void Dispatcher::create_context(QOpenGLContext *context)
{
    m_uploader.create_context(context);
}

void Dispatcher::open(QString host, int port)
{
    QByteArray tmp = host.toLocal8Bit();
    m_client = rfbGetClient(8, 3, 4);

    m_client->canHandleNewFBSize = FALSE;
    m_client->serverHost = new char[tmp.size() + 1];
    strcpy(m_client->serverHost, tmp.constData());
    m_client->serverPort = port;
    m_client->MallocFrameBuffer = Dispatcher::MallocFrameBuffer;
    m_client->GotCopyRect = Dispatcher::GotCopyRect;
    m_client->GotFillRect = Dispatcher::GotFillRect;
    m_client->GotBitmap = Dispatcher::GotBitmap;
    m_client->GotJpeg = Dispatcher::GotJpeg;
    m_client->FinishedFrameBufferUpdate = Dispatcher::FinishedFrameBufferUpdate;

    rfbClientSetClientData(m_client, 0, this);

    if(!rfbInitClient(m_client, NULL, NULL)) {
        emit error();
        return;
    }

    refresh();
}
void Dispatcher::refresh()
{
    if(m_client == NULL)
        return;

    setFinished(false);
    m_uploader.startedUpdate();
    int message = 0;

    // Wait until we get something
    while(message == 0) {
        message = WaitForMessage(m_client, 10000);
        if(message < 0) {
            terminate();
            emit error();
            return;
        }
        if (message) {
            if(!HandleRFBServerMessage(m_client)) {
                terminate();
                emit error();
                return;
            }
        }
    }
}

void Dispatcher::terminate()
{
    if(m_client == NULL)
        return;

    rfbClientCleanup(m_client);
}

void Dispatcher::decodingFinished(const QImage &bitmap, const int x, const int y, const int w, const int h)
{
    m_uploader.gotBitmap(bitmap, x, y, w, h);
    if(m_finished && decrementDecoderCount() == 0) {
        m_uploader.finishedUpdate();
        emit updateFramebuffer();
    }
}
