#include "vncrenderer.h"

VncRenderer::VncRenderer(QQuickWindow *window,
                         QOpenGLBuffer fill_buffer,
                         QSharedPointer<QVector<GLuint>> copy_textures,
                         QOpenGLBuffer copy_buffer,
                         QOpenGLBuffer draw_buffer,
                         QSharedPointer<QVector<QOpenGLTexture*>> draw_textures,
                         QObject *parent) :
    QObject(parent), m_window(window), m_clear(true), m_fill_buffer(fill_buffer), m_copy_textures(copy_textures),
    m_copy_buffer(copy_buffer), m_draw_buffer(draw_buffer), m_draw_textures(draw_textures)
{
    m_fill_program.addShaderFromSourceFile(QOpenGLShader::Vertex, "fill_shader.vsh");
    m_fill_program.addShaderFromSourceFile(QOpenGLShader::Fragment, "fill_shader.fsh");
    m_fill_program.link();

    m_draw_program.addShaderFromSourceFile(QOpenGLShader::Vertex, "draw_shader.vsh");
    m_draw_program.addShaderFromSourceFile(QOpenGLShader::Fragment, "draw_shader.fsh");
    m_draw_program.bind();
    m_draw_program.setUniformValue("texture", GL_TEXTURE_2D);
    m_draw_program.release();

    m_functions = QOpenGLContext::currentContext()->functions();
}

void VncRenderer::render()
{
    if(m_clear) {
        m_functions->glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        m_functions->glClear(GL_COLOR_BUFFER_BIT);
        m_clear = false;
    }

    if(m_fill_count > 0) {
        m_fill_program.bind();
        m_fill_program.enableAttributeArray("position");
        m_fill_program.enableAttributeArray("color");
        m_fill_buffer.bind();
        m_fill_program.setAttributeBuffer("position", GL_FLOAT, 0 * sizeof(float), 2, 5 * sizeof(float));
        m_fill_program.setAttributeBuffer("color", GL_FLOAT, 2 * sizeof(float), 3, 5 * sizeof(float));
        for(unsigned int i = 0; i < m_fill_count; i++)
            m_functions->glDrawArrays(GL_TRIANGLE_STRIP, i * 4, 4);
        m_fill_buffer.release();
        m_fill_program.release();
    }

    if(m_copy_count > 0 || m_draw_count > 0)
        m_draw_program.bind();
        m_draw_program.enableAttributeArray("position");
        m_draw_program.enableAttributeArray("texcoord");

    if(m_copy_count > 0){
        m_copy_buffer.bind();
        m_draw_program.setAttributeBuffer("position", GL_FLOAT, 0 * sizeof(float), 2, 4 * sizeof(float));
        m_draw_program.setAttributeBuffer("texcoord", GL_FLOAT, 2 * sizeof(float), 2, 4 * sizeof(float));

        for(unsigned int i = 0; i < m_copy_count; i++) {
            m_functions->glBindTexture(GL_TEXTURE_2D, m_copy_textures->at(i));
            m_functions->glDrawArrays(GL_TRIANGLE_STRIP, i * 4, 4);
        }
        m_functions->glBindTexture(GL_TEXTURE_2D, 0);
        m_copy_buffer.release();
    }

    if(m_draw_count > 0) {
        m_draw_buffer.bind();
        m_draw_program.setAttributeBuffer("position", GL_FLOAT, 0 * sizeof(float), 2, 4 * sizeof(float));
        m_draw_program.setAttributeBuffer("texcoord", GL_FLOAT, 2 * sizeof(float), 2, 4 * sizeof(float));

        for(unsigned int i = 0; i < m_draw_count; i++) {
            m_draw_textures->at(i)->bind();
            m_functions->glDrawArrays(GL_TRIANGLE_STRIP, i * 4, 4);
            m_draw_textures->at(i)->release();
        }

        m_draw_buffer.release();
    }

    if(m_copy_count > 0 || m_draw_count > 0)
        m_draw_program.release();

    m_window->resetOpenGLState();

    emit finishedRendering();
}

QOpenGLFramebufferObject *VncRenderer::createFramebufferObject(const QSize &size)
{
    QMatrix4x4 ortho;
    ortho.ortho(QRect(QPoint(0, size.height()), QSize(size.width(), -1 * size.height())));
    m_fill_program.bind();
    m_fill_program.setUniformValue("ortho", ortho);
    m_draw_program.bind();
    m_draw_program.setUniformValue("ortho", ortho);
    m_draw_program.release();

    QOpenGLFramebufferObject *framebufferObject = new QOpenGLFramebufferObject(size);
    emit FBOTextureChanged(framebufferObject->texture(), ortho);
    return framebufferObject;
}

void VncRenderer::synchronize(QQuickFramebufferObject *object)
{
    QVnc *vnc = static_cast<QVnc*>(object);
    Dispatcher *dispatcher = vnc->dispatcher();
    m_fill_count = dispatcher->fill_count();
    m_copy_count = dispatcher->copy_count();
    m_draw_count = dispatcher->draw_count();

    m_fill_program.bind();
    m_fill_program.setUniformValue("colormax", dispatcher->colorMax());
    m_fill_program.release();
}
